/*
 **
 ** Copyright 2019 Karlsruhe Institute of Technology, Chair of Dependable Nano Computing
 **
 ** Permission is hereby granted, free of charge, to any person obtaining a copy of this software
 ** and associated documentation files (the "Software"), to deal in the Software without restriction, 
 ** including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 ** and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do 
 ** so, subject to the following conditions:
 **
 ** The above copyright notice and this permission notice shall be included in all copies or substantial 
 ** portions of the Software.
 **
 ** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT 
 ** LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 ** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
 ** WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
 ** SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 **
 */
module crash (clkin, rstin, uart_rxin, uart_txout, ledout);
	
parameter amount = 7100;

input clkin;
input rstin;
input uart_rxin;
output uart_txout;
output [7:0] ledout;

wire rst;
wire clk60mhz;

reg [27:0] ctr60= 0;
reg [25:0] ctr12 = 0;
reg ctr60led = 1'b0;
reg ctr12led = 1'b0;

assign ledout[0] = ctr12led;
assign ledout[1] = ctr60led;
assign ledout[3] = ena_toggle;

reg ena_toggle = 1'b0;
reg [20:0] ena_ctr;
reg [20:0] cycles = 16;
reg [20:0] activecycles  = 1;
reg [25:0] togglectr;
reg [1:0] enabuf = 2'b0;
reg crash_ena = 1'b0;

wire [7:0] uart_rxdata;
wire uart_rxrdy;

debounce rstin_debounce(.clkin(clkin), .asynchin(rstin), .synchout(rst));
pllc pll_inst(.clock_in(clkin), .reset_in(1'b1), .clock_out(clk60mhz));
uart uart_inst(.clkin(clkin), .rstin(rst), .txdatain(8'b0), .txrdyin(1'b0), .rxin(uart_rxin), .rxdataout(uart_rxdata), 
				.rxrdyout(uart_rxrdy), .txout(uart_txout), .errout(ledout[7]));
defparam uart_inst.CLKS_PER_BIT = 625;

// synthesis loop_limit 8000
genvar i;
generate
	for (i = 0; i < amount; i = i+1) begin : ros
		ringosc osc(.enable(ena_toggle)) /* synthesis syn_noprune=1 syn_preserve=1 */;
	end
endgenerate

always @(posedge clk60mhz, negedge rst) begin
	if (!rst) begin
		ctr60 <= 0;
	end else if (ctr60 == 60000000) begin
		ctr60 <= 0;
		ctr60led <= ~ctr60led;
	end else begin
		ctr60 <= ctr60 + 1;
	end
end


always @(posedge clkin, negedge rst) begin
	if (!rst) begin
		ctr12 <= 0;
	end else if (ctr12 == 12000000) begin
		ctr12 <= 0;
		ctr12led <= ~ctr12led;
	end else begin
		ctr12 <= ctr12 + 1;
	end
end

always @(posedge clk60mhz, negedge rst) begin
	if (!rst) begin
		togglectr <= 0;
		cycles <= 16;
		activecycles <= 1;
	end else if (togglectr >= 5000) begin
		togglectr <= 0;
		if (activecycles < ((cycles >> 1) + (cycles >> 2))) begin
			activecycles <= activecycles + (cycles >> 4);
		end else begin
			cycles <= cycles + 128;
			activecycles <= ((cycles + 128) >>  4);
		end
	end else begin
		togglectr <= togglectr + 1'b1;
	end
end

always @(posedge clk60mhz, negedge rst) begin
	if (!rst) begin
		crash_ena <= 1'b0;
	end else if ((uart_rxrdy) && (uart_rxdata == 8'h73)) begin
		crash_ena <= 1'b1;
	end
end

always @(posedge clk60mhz, negedge rst) begin
	if (!rst) begin
		ena_ctr <= 0;
	end else if (crash_ena && (ena_ctr >= cycles)) begin
		ena_ctr <= 0;
	end else if (crash_ena) begin
		ena_ctr <= ena_ctr + 1'b1;
	end
end

always @(posedge clk60mhz, negedge rst) begin
	if (!rst) begin
		ena_toggle <= 1'b0;
	end else if (crash_ena && (ena_ctr <= activecycles)) begin
		ena_toggle <= 1'b1;
	end else begin
		ena_toggle <= 1'b0;
	end
end

endmodule


