/*
 **
 ** Copyright 2019 Karlsruhe Institute of Technology, Chair of Dependable Nano Computing
 **
 ** Permission is hereby granted, free of charge, to any person obtaining a copy of this software
 ** and associated documentation files (the "Software"), to deal in the Software without restriction, 
 ** including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 ** and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do 
 ** so, subject to the following conditions:
 **
 ** The above copyright notice and this permission notice shall be included in all copies or substantial 
 ** portions of the Software.
 **
 ** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT 
 ** LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 ** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
 ** WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
 ** SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 **
 */

#include "AESCracker.h"

AESCracker::AESCracker() {
	uint8_t c = 0x63;
	exp[0] = 0x01;
	log[0] = 0x00;

	for (uint8_t i = 0x01; i > 0x00; i++) {
		exp[i] = mult(exp[i-1], 0x03);
		log[exp[i]] = i;

	}

	sbox[0x00] = 0x63;
	invsbox[0x63] = 0x00;
	for (uint8_t i = 0x01; i > 0x00; i++) {
		uint8_t j = inv(i);
		uint8_t s = j ^ ror(j, 4) ^ ror(j, 5) ^ ror(j, 6) ^ ror(j, 7) ^ c;
		invsbox[s] = i;
		sbox[i] = s;
	}

	uint8_t x = 0x01;
	for (unsigned int i = 1; i < 11; i++) {
		rcon[i] = x;
        x = ((x << 1) ^ ((x & 0x80) ? 0x1b : 0x00)) & 0xff;
	}

	unsigned int shift = 24;
	uint32_t tmpdiff = 0x00000000;

	for (unsigned int i = 0; i < 16; i++) {
		for (unsigned int j = 0; j < 16; j++) {
			state[j%4][j/4] = 0x00;
		}
		shiftRows();
		mixColumns();
		for (unsigned int j = 0; j < 16; j++) {
			if (state[j%4][j/4] != 0x00) {
                tmpdiff |= static_cast<uint32_t>((state[j%4][j/4]) << shift);
				shift -= 8;
			}
		}
        unsigned int type = 0;
		switch (i) {
		case  0:
		case  5:
		case 10:
		case 15: type = 0;
		break;
		case  3:
		case  4:
		case  9:
		case 14: type = 1;
		break;
		case  2:
		case  7:
		case  8:
		case 13: type = 2;
		break;
		case  1:
		case  6:
		case 11:
		case 12: type = 3;
		break;
		}
		diff[type].insert(tmpdiff);
		diff2byte[type].insert(tmpdiff & 0xFFFF0000);
		diff3byte[type].insert(tmpdiff & 0xFFFFFF00);
		for (uint8_t x = 0x01; x != 0x00; x++) {
			for (unsigned int j = 0; j < 16; j++) {
				state[j%4][j/4] = (i == j) ? (x) : (0x00);
			}
			shiftRows();
			mixColumns();
			shiftRows();
			shift = 24;
			tmpdiff = 0x00000000;
			for (unsigned int j = 0; j < 16; j++) {
				if (state[j%4][j/4] != 0x00) {
                    tmpdiff |= static_cast<uint32_t>((state[j%4][j/4]) << shift);
					shift -= 8;
				}
			}
			diff[type].insert(tmpdiff);
			diff2byte[type].insert(tmpdiff & 0xFFFF0000);
			diff3byte[type].insert(tmpdiff & 0xFFFFFF00);
		}
	}
}

unsigned int AESCracker::getTotalLog2Candidates() const {
	unsigned int total = 0;
	for (uint8_t i = 0; i < 0x04; i++) {
		if (key_candidates[i].size() == 0) {
			total += 32;
		} else {
			total += std::log2(key_candidates[i].size());
		}
	}
	return total;
}

unsigned int AESCracker::getTotalNumCandidates() const {
	unsigned int total = 1;
	for (uint8_t i = 0; i < 0x04; i++) {
		if (key_candidates[i].size() == 0) {
			total = 0xFFFFFFFF;
		} else {
			total *= key_candidates[i].size();
		}
	}
	return total;
}

int AESCracker::determineFaultType(const Block &correct, const Block &faulty) const {
	int type = -1;
	std::list<uint8_t> pos;
	for (uint8_t i = 0x00; i < 0x10; i++) {
		if (correct[i] != faulty[i]) {
			pos.push_back(i);
		}
	}
	if (pos.size() == 4) {
		pos.sort();
		uint32_t poscode = 0x00;
		for (unsigned int i = 0; i < 4; i++) {
            poscode |= static_cast<uint32_t>((pos.back()) << (i*8));
			pos.pop_back();
		}
		switch (poscode) {
		case 0x00070A0D: type = 0;
							break;
		case 0x01040B0E: type = 1;
							break;
		case 0x0205080F: type = 2;
							break;
		case 0x0306090C: type = 3;
							break;
		default: type = -1;
		}
	}
	return type;
}

uint32_t AESCracker::invSubBytesWord(uint32_t w) const {
	uint32_t tmp = 0x00000000;
    tmp |= static_cast<uint32_t>(invsbox[static_cast<uint8_t>(w & 0xFF)]);
    tmp |= static_cast<uint32_t>(invsbox[static_cast<uint8_t>((w >> 8) & 0xFF)] << 8);
    tmp |= static_cast<uint32_t>(invsbox[static_cast<uint8_t>((w >> 16) & 0xFF)] << 16);
    tmp |= static_cast<uint32_t>(invsbox[static_cast<uint8_t>((w >> 24) & 0xFF)] << 24);
	return tmp;
}

uint32_t AESCracker::computeDifference(uint32_t key_candidate, unsigned int faulttype, const Block &correct, const Block &faulty) {
	uint32_t correctword = 0x00000000;
	uint32_t faultyword = 0x00000000;
	switch (faulttype) {
	case 0:
        correctword |= static_cast<uint32_t>(correct[0] << 24);
        correctword |= static_cast<uint32_t>(correct[7] << 16);
        correctword |= static_cast<uint32_t>(correct[10] << 8);
        correctword |= static_cast<uint32_t>(correct[13]);
        faultyword |= static_cast<uint32_t>(faulty[0] << 24);
        faultyword |= static_cast<uint32_t>(faulty[7] << 16);
        faultyword |= static_cast<uint32_t>(faulty[10] << 8);
        faultyword |= static_cast<uint32_t>(faulty[13]);
	break;
	case 1:
        correctword |= static_cast<uint32_t>(correct[1] << 24);
        correctword |= static_cast<uint32_t>(correct[4] << 16);
        correctword |= static_cast<uint32_t>(correct[11] << 8);
        correctword |= static_cast<uint32_t>(correct[14]);
        faultyword |= static_cast<uint32_t>(faulty[1] << 24);
        faultyword |= static_cast<uint32_t>(faulty[4] << 16);
        faultyword |= static_cast<uint32_t>(faulty[11] << 8);
        faultyword |= static_cast<uint32_t>(faulty[14]);
	break;
	case 2:
        correctword |= static_cast<uint32_t>(correct[2] << 24);
        correctword |= static_cast<uint32_t>(correct[5] << 16);
        correctword |= static_cast<uint32_t>(correct[8] << 8);
        correctword |= static_cast<uint32_t>(correct[15]);
        faultyword |= static_cast<uint32_t>(faulty[2] << 24);
        faultyword |= static_cast<uint32_t>(faulty[5] << 16);
        faultyword |= static_cast<uint32_t>(faulty[8] << 8);
        faultyword |= static_cast<uint32_t>(faulty[15]);
	break;
	case 3:
        correctword |= static_cast<uint32_t>(correct[3] << 24);
        correctword |= static_cast<uint32_t>(correct[6] << 16);
        correctword |= static_cast<uint32_t>(correct[9] << 8);
        correctword |= static_cast<uint32_t>(correct[12]);
        faultyword |= static_cast<uint32_t>(faulty[3] << 24);
        faultyword |= static_cast<uint32_t>(faulty[6] << 16);
        faultyword |= static_cast<uint32_t>(faulty[9] << 8);
        faultyword |= static_cast<uint32_t>(faulty[12]);
	break;
	}

	return (invSubBytesWord(correctword ^ key_candidate) ^ invSubBytesWord(faultyword ^ key_candidate));
}

void AESCracker::filterCandidates(unsigned int faulttype, const Block &correct, const Block &faulty) {
	std::list<uint32_t>::iterator it = key_candidates[faulttype].begin();
	while (it != key_candidates[faulttype].end()) {
		uint32_t tmpdiff = computeDifference(*it, faulttype, correct, faulty);
		if (diff[faulttype].find(tmpdiff) == diff[faulttype].end()) {
			it = key_candidates[faulttype].erase(it);
		} else {
			it++;
		}
	}
}

void AESCracker::precomputeCandidates(unsigned int faulttype) {
	std::list<uint32_t> candidates2byte;
	for (uint32_t x = 0x00000000; x != 0x00010000; x++) {
			uint32_t firstdiff = 0xFFFF0000 & computeDifference(x << 16, faulttype, precomputation_pairs[faulttype].front().first, precomputation_pairs[faulttype].front().second);
			uint32_t seconddiff = 0xFFFF0000 & computeDifference(x << 16, faulttype, precomputation_pairs[faulttype].back().first, precomputation_pairs[faulttype].back().second);
			if ((diff2byte[faulttype].find(firstdiff) != diff2byte[faulttype].end()) && (diff2byte[faulttype].find(seconddiff) != diff2byte[faulttype].end())) {
				candidates2byte.push_back(x << 16);
			}
	}
	std::list<uint32_t> candidates3byte;
	for (uint32_t x: candidates2byte) {
		uint32_t firstdiff = 0xFFFFFF00 & computeDifference(x, faulttype, precomputation_pairs[faulttype].front().first, precomputation_pairs[faulttype].front().second);
		uint32_t seconddiff = 0xFFFFFF00 & computeDifference(x, faulttype, precomputation_pairs[faulttype].back().first, precomputation_pairs[faulttype].back().second);
		if ((diff2byte[faulttype].find(firstdiff) != diff2byte[faulttype].end()) && (diff2byte[faulttype].find(seconddiff) != diff2byte[faulttype].end())) {
			candidates3byte.push_back(x);
		}
		for (uint8_t y = 0x01; y != 0x00; y++) {
            uint32_t firstdiff = 0xFFFFFF00 & computeDifference(x | static_cast<uint32_t>(y << 8), faulttype, precomputation_pairs[faulttype].front().first, precomputation_pairs[faulttype].front().second);
            uint32_t seconddiff = 0xFFFFFF00 & computeDifference(x | static_cast<uint32_t>(y << 8), faulttype, precomputation_pairs[faulttype].back().first, precomputation_pairs[faulttype].back().second);
			if ((diff3byte[faulttype].find(firstdiff) != diff3byte[faulttype].end()) && (diff3byte[faulttype].find(seconddiff) != diff3byte[faulttype].end())) {
                    candidates3byte.push_back(x | static_cast<uint32_t>(y << 8));
			}
		}
	}
	for (uint32_t x: candidates3byte) {
		uint32_t firstdiff = computeDifference(x, faulttype, precomputation_pairs[faulttype].front().first, precomputation_pairs[faulttype].front().second);
		uint32_t seconddiff = computeDifference(x, faulttype, precomputation_pairs[faulttype].back().first, precomputation_pairs[faulttype].back().second);
		if ((diff[faulttype].find(firstdiff) != diff[faulttype].end()) && (diff[faulttype].find(seconddiff) != diff[faulttype].end())) {
			key_candidates[faulttype].push_back(x);
		}
		for (uint8_t y = 0x01; y != 0x00; y++) {
			uint32_t firstdiff = computeDifference(x | y, faulttype, precomputation_pairs[faulttype].front().first, precomputation_pairs[faulttype].front().second);
			uint32_t seconddiff = computeDifference(x | y, faulttype, precomputation_pairs[faulttype].back().first, precomputation_pairs[faulttype].back().second);
			if ((diff[faulttype].find(firstdiff) != diff[faulttype].end()) && (diff[faulttype].find(seconddiff) != diff[faulttype].end())) {
				key_candidates[faulttype].push_back(x | y);
			}
		}
	}
}

int AESCracker::evaluatePair(const Block &correct, const Block &faulty) {
	int type = determineFaultType(correct, faulty);
	if (type >= 0) {
        unsigned int type_index = static_cast<unsigned int>(type);
        if (key_candidates[type_index].size() == 0) {
            precomputation_pairs[type_index].push_back(std::make_pair(correct, faulty));
            if (precomputation_pairs[type_index].size() < 2) {
				return -2;
            } else if (precomputation_pairs[type_index].size() > 2) {
                precomputation_pairs[type_index].pop_front();
                precomputeCandidates(type_index);
            } else {
                precomputeCandidates(type_index);
			}
            for (auto pair: precomputation_pairs[type_index]) {
                filterCandidates(type_index, pair.first, pair.second);
			}
        } else if (key_candidates[type_index].size() != 1) {
            filterCandidates(type_index, correct, faulty);
            if (key_candidates[type_index].size() == 0) {
                precomputation_pairs[type_index].clear();
            }
		}
        return 1;
	} else {
		return -1;
	}
}

bool AESCracker::isFeasibile(const Block &correct, const Block &faulty) const {
	return (determineFaultType(correct, faulty) >= 0);
}

void AESCracker::reset() {
    std::fill(std::begin(key_candidates), std::end(key_candidates), std::list<uint32_t>());
    std::fill(std::begin(precomputation_pairs), std::end(precomputation_pairs), std::list<std::pair<Block, Block>>());
}

Block AESCracker::getRoundKeyCandidate() const {
    Block candidate;
	if (key_candidates[0].size() > 0) {
		uint32_t tmp = key_candidates[0].front();
        candidate[0] = (tmp >> 24) & 0xff;
        candidate[7] = (tmp >> 16) & 0xff;
        candidate[10] = (tmp >> 8) & 0xff;
        candidate[13] = tmp & 0xff;
	}
	if (key_candidates[1].size() > 0) {
		uint32_t tmp = key_candidates[1].front();
        candidate[1] = (tmp >> 24) & 0xff;
        candidate[4] = (tmp >> 16) & 0xff;
        candidate[11] = (tmp >> 8) & 0xff;
        candidate[14] = tmp & 0xff;
	}
	if (key_candidates[2].size() > 0) {
		uint32_t tmp = key_candidates[2].front();
        candidate[2] = (tmp >> 24) & 0xff;
        candidate[5] = (tmp >> 16) & 0xff;
        candidate[8] = (tmp >> 8) & 0xff;
        candidate[15] = tmp & 0xff;
	}
	if (key_candidates[3].size() > 0) {
		uint32_t tmp = key_candidates[3].front();
        candidate[3] = (tmp >> 24) & 0xff;
        candidate[6] = (tmp >> 16) & 0xff;
        candidate[9] = (tmp >> 8) & 0xff;
        candidate[12] = tmp & 0xff;
	}
	return candidate;
}

Block AESCracker::recoverAESKey(const Block &roundkey) const {
	std::array<uint8_t, 176> roundkeys;
	for (unsigned int i = 0; i < 16; i++) {
		roundkeys[160 + i] = roundkey[i];
	}

    for (unsigned int i = 43; i > 3; i--) {
		std::array<uint8_t, 4> temp = {roundkeys[(i-1)*4], roundkeys[(i-1)*4 + 1], roundkeys[(i-1)*4 + 2], roundkeys[(i-1)*4 + 3]};
		if (i%4 == 0) {
			std::array<uint8_t, 4> oldtemp(temp);
			temp[0] = sbox[oldtemp[1]] ^ rcon[i/4];
			temp[1] = sbox[oldtemp[2]];
			temp[2] = sbox[oldtemp[3]];
			temp[3] = sbox[oldtemp[0]];
		}
		roundkeys[(i-4)*4] = roundkeys[4*i] ^ temp[0];
		roundkeys[(i-4)*4 + 1] = roundkeys[4*i + 1] ^ temp[1];
		roundkeys[(i-4)*4 + 2] = roundkeys[4*i + 2] ^ temp[2];
		roundkeys[(i-4)*4 + 3] = roundkeys[4*i + 3] ^ temp[3];
	}

    Block aeskey;
	for (unsigned int i = 0; i < 16; i++) {
		aeskey[i] = roundkeys[i];
	}
	return aeskey;
}

void AESCracker::shiftRow(unsigned int i, int j) {
	std::array<uint8_t, 4> copy(state[i]);
    for (unsigned int k = 0; k < 4; k++) {
        state[i][k] = copy[(static_cast<int>(k)+j) % 4];
	}
}

void AESCracker::shiftRows() {
    for (int i = 0; i < 4; i++) {
        shiftRow(static_cast<unsigned int>(i), i);
	}
}

void AESCracker::invShiftRows() {
    for (int i = 0; i < 4; i++) {
        shiftRow(static_cast<unsigned int>(i), -i);
	}
}

void AESCracker::mixColumns() {
	std::array<uint8_t,4> tempcolumn;
	for (unsigned int i = 0; i < 4; i++) {
		tempcolumn[0] = fast_mult(state[0][i], 0x02) ^ fast_mult(state[1][i], 0x03) ^ state[2][i] ^ state[3][i];
		tempcolumn[1] = fast_mult(state[1][i], 0x02) ^ fast_mult(state[2][i], 0x03) ^ state[0][i] ^ state[3][i];
		tempcolumn[2] = fast_mult(state[2][i], 0x02) ^ fast_mult(state[3][i], 0x03) ^ state[0][i] ^ state[1][i];
		tempcolumn[3] = fast_mult(state[3][i], 0x02) ^ fast_mult(state[0][i], 0x03) ^ state[1][i] ^ state[2][i];
		for (unsigned int j = 0; j < 4; j++) {
			state[j][i] = tempcolumn[j];
		}
	}
}

void AESCracker::invMixColumns() {
	std::array<uint8_t,4> tempcolumn;
	for (unsigned int i = 0; i < 4; i++) {
		tempcolumn[0] = fast_mult(state[0][i], 0x0E) ^ fast_mult(state[1][i], 0x0B) ^ fast_mult(state[2][i], 0x0D) ^ fast_mult(state[3][i], 0x09);
		tempcolumn[1] = fast_mult(state[0][i], 0x09) ^ fast_mult(state[1][i], 0x0E) ^ fast_mult(state[2][i], 0x0B) ^ fast_mult(state[3][i], 0x0D);
		tempcolumn[2] = fast_mult(state[0][i], 0x0D) ^ fast_mult(state[1][i], 0x09) ^ fast_mult(state[2][i], 0x0E) ^ fast_mult(state[3][i], 0x0B);
		tempcolumn[3] = fast_mult(state[0][i], 0x0B) ^ fast_mult(state[1][i], 0x0D) ^ fast_mult(state[2][i], 0x09) ^ fast_mult(state[3][i], 0x0E);
		for (unsigned int j = 0; j < 4; j++) {
			state[j][i] = tempcolumn[j];
		}
	}
}

void AESCracker::invSubBytes() {
	for (unsigned int i = 0; i < 4; i++) {
		for (unsigned int j = 0; j < 4; j++) {
			state[i][j] = invsbox[state[i][j]];
		}
	}
}

AESCracker::~AESCracker() {
	// TODO Auto-generated destructor stub
}

