/*
 **
 ** Copyright 2019 Karlsruhe Institute of Technology, Chair of Dependable Nano Computing
 **
 ** Permission is hereby granted, free of charge, to any person obtaining a copy of this software
 ** and associated documentation files (the "Software"), to deal in the Software without restriction, 
 ** including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 ** and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do 
 ** so, subject to the following conditions:
 **
 ** The above copyright notice and this permission notice shall be included in all copies or substantial 
 ** portions of the Software.
 **
 ** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT 
 ** LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 ** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
 ** WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
 ** SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 **
 */
#ifndef AESCRACKER_H_
#define AESCRACKER_H_

#include <array>
#include <unordered_set>
#include <list>
#include <cinttypes>
#include <utility>
#include <cmath>
#include "types.h"

class AESCracker {
private:

	std::array<std::array<uint8_t, 4>, 4> state;
	std::array<uint8_t, 256> sbox;
	std::array<uint8_t, 256> invsbox;
	std::array<uint8_t, 256> log;
	std::array<uint8_t, 256> exp;
	std::array<uint8_t, 11> rcon;
	std::array<std::list<uint32_t>, 4> key_candidates;
    std::array<std::list<std::pair<Block, Block>>, 4> precomputation_pairs;
	std::array<std::unordered_set<uint32_t>, 4>  diff;
	std::array<std::unordered_set<uint32_t>, 4>  diff2byte;
	std::array<std::unordered_set<uint32_t>, 4>  diff3byte;
	unsigned int round = 0;

	inline uint8_t ror(uint8_t x, unsigned int m) const {
        return ((x >> m) | (x << (8 - m))) & 0xff;
	}
	inline uint8_t inv(uint8_t x) const {
		return (x ? exp[255 - log[x]] : 0x00);
	}
	inline uint8_t mult(uint8_t a, uint8_t b) const {
		uint8_t product = 0;
		for (unsigned int i = 0; i < 8; i++) {
			product ^= (1 & b) ? a : 0x00;
            a = ((a << 1) ^ ((a & 0x80) ? 0x1b : 0x00)) & 0xff;
			b >>= 1;
		}
		return product;
	}
	inline uint8_t fast_mult(uint8_t a, uint8_t b) const {
		return ((a && b) ? exp[(log[a] + log[b]) % 255] : 0x00);
	}

	void shiftRow(unsigned int i, int j);
	void shiftRows();
	void invShiftRows();
	void mixColumns();
	void invMixColumns();
	void subBytes();
	void invSubBytes();
	uint32_t invSubBytesWord(uint32_t w) const;
    uint32_t computeDifference(uint32_t key_candidate, unsigned int faulttype, const Block &correct, const Block &faulty);
	void precomputeCandidates(unsigned int faulttype);
    void filterCandidates(unsigned int faulttype, const Block &correct, const Block &faulty);

public:
	AESCracker();
	unsigned int getTotalNumCandidates() const;
    unsigned int getTotalLog2Candidates() const;
    bool isFeasibile(const Block &correct, const Block &faulty) const;
    int determineFaultType(const Block &correct, const Block &faulty) const;
	void reset();
    Block getRoundKeyCandidate() const;
    Block recoverAESKey(const Block &roundkey) const;
    int evaluatePair(const Block &correct, const Block &faulty);
	virtual ~AESCracker();
};

#endif /* AESCRACKER_H_ */
