/*
 **
 ** Copyright 2019 Karlsruhe Institute of Technology, Chair of Dependable Nano Computing
 **
 ** Permission is hereby granted, free of charge, to any person obtaining a copy of this software
 ** and associated documentation files (the "Software"), to deal in the Software without restriction, 
 ** including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 ** and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do 
 ** so, subject to the following conditions:
 **
 ** The above copyright notice and this permission notice shall be included in all copies or substantial 
 ** portions of the Software.
 **
 ** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT 
 ** LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 ** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
 ** WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
 ** SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 **
 */

#ifndef AESHELPER_H_
#define AESHELPER_H_

#include <array>
#include "types.h"

class AESHelper {
private:

	std::array<uint8_t, 256> sbox;
	std::array<uint32_t, 256> t1;
	std::array<uint32_t, 256> t2;
	std::array<uint32_t, 256> t3;
	std::array<uint32_t, 256> t4;
	std::array<uint8_t, 256> invsbox;
	std::array<uint8_t, 256> log;
	std::array<uint8_t, 256> exp;
	std::array<uint8_t, 11> rcon;
	std::array<uint8_t, 176> roundkeys;

public:
    AESHelper(const Block &key) {
		uint8_t c = 0x63;
		exp[0] = 0x01;
		log[0] = 0x00;

		for (uint8_t i = 0x01; i > 0x00; i++) {
			exp[i] = mult(exp[i-1], 0x03);
			log[exp[i]] = i;
		}

		sbox[0] = 0x63;
		invsbox[0x63] = 0x00;
		for (uint8_t i = 0x01; i > 0x00; i++) {
			uint8_t j = inv(i);
			uint8_t s = j ^ ror(j, 4) ^ ror(j, 5) ^ ror(j, 6) ^ ror(j, 7) ^ c;
			sbox[i] = s;
			invsbox[s] = i;
		}
        t1[0] = static_cast<uint32_t>(0x00000000 | (fast_mult(2, sbox[0]) | (sbox[0] << 8) | (sbox[0] << 16) | (fast_mult(3, sbox[0]) << 24)));
        t2[0] = static_cast<uint32_t>(0x00000000 | (fast_mult(3, sbox[0]) | (fast_mult(2, sbox[0]) << 8) | (sbox[0] << 16) | (sbox[0] << 24)));
        t3[0] = static_cast<uint32_t>(0x00000000 | (sbox[0] | (fast_mult(3, sbox[0]) << 8) | (fast_mult(2, sbox[0]) << 16) | (sbox[0] << 24)));
        t4[0] = static_cast<uint32_t>(0x00000000 | (sbox[0] | (sbox[0] << 8) | (fast_mult(3, sbox[0]) << 16) | (fast_mult(2, sbox[0]) << 24)));
		for (uint8_t i = 0x01; i > 0x00; i++) {
            t1[i] = static_cast<uint32_t>(0x00000000 | (fast_mult(2, sbox[i]) | (sbox[i] << 8) | (sbox[i] << 16) | (fast_mult(3, sbox[i]) << 24)));
            t2[i] = static_cast<uint32_t>(0x00000000 | (fast_mult(3, sbox[i]) | (fast_mult(2, sbox[i]) << 8) | (sbox[i] << 16) | (sbox[i] << 24)));
            t3[i] = static_cast<uint32_t>(0x00000000 | (sbox[i] | (fast_mult(3, sbox[i]) << 8) | (fast_mult(2, sbox[i]) << 16) | (sbox[i] << 24)));
            t4[i] = static_cast<uint32_t>(0x00000000 | (sbox[i] | (sbox[i] << 8) | (fast_mult(3, sbox[i]) << 16) | (fast_mult(2, sbox[i]) << 24)));
		}
		uint8_t x = 0x01;
		for (unsigned int i = 1; i < 11; i++) {
			rcon[i] = x;
            x = ((x << 1) ^ ((x & 0x80) ? 0x1b : 0x00)) & 0xff;
		}
		for (unsigned int i = 0; i < 16; i++) {
            roundkeys[i] = key[i];
		}
		for (unsigned int i = 4; i < 44; i++) {
			std::array<uint8_t, 4> temp = {roundkeys[(i-1)*4], roundkeys[(i-1)*4 + 1], roundkeys[(i-1)*4 + 2], roundkeys[(i-1)*4 + 3]};
			if ((i % 4) == 0) {
				std::array<uint8_t, 4> oldtemp(temp);
				temp[0] = sbox[oldtemp[1]] ^ rcon[i/4];
				temp[1] = sbox[oldtemp[2]];
				temp[2] = sbox[oldtemp[3]];
				temp[3] = sbox[oldtemp[0]];
			}
			roundkeys[4*i] = temp[0] ^ roundkeys[(i-4)*4];
			roundkeys[4*i + 1] = temp[1] ^ roundkeys[(i-4)*4 + 1];
			roundkeys[4*i + 2] = temp[2] ^ roundkeys[(i-4)*4 + 2];
			roundkeys[4*i + 3] = temp[3] ^ roundkeys[(i-4)*4 + 3];
		}
	}

	inline uint8_t ror(uint8_t x, unsigned int m) const {
        return ((x >> m) | (x << (8 - m))) & 0xff;
	}

	inline uint8_t inv(uint8_t x) const {
		return (x ? exp[255 - log[x]] : 0x00);
	}

	inline uint8_t mult(uint8_t a, uint8_t b) const {
		uint8_t product = 0;
		for (unsigned int i = 0; i < 8; i++) {
			product ^= (1 & b) ? a : 0x00;
            a = ((a << 1) ^ ((a & 0x80) ? 0x1b : 0x00)) & 0xff;
			b >>= 1;
		}
		return product;
	}

	inline uint8_t fast_mult(uint8_t a, uint8_t b) const {
		return ((a && b) ? exp[(log[a] + log[b]) % 255] : 0x00);
	}

	inline uint8_t sbox_val(uint8_t in) const {
		return sbox[in];
	}

	inline uint32_t t_val(uint8_t in, unsigned int byteindex) const {
		switch(byteindex) {
		case 0:
		case 4:
		case 8:
		case 12: 	return t1[in];
		case 5:
		case 9:
		case 13:
		case 1: 	return t2[in];
		case 10:
		case 14:
		case 2:
		case 6: 	return t3[in];
		case 15:
		case 3:
		case 7:
		case 11: 	return t4[in];
		default:
					return t1[in];
		}
	}

	inline uint8_t isbox_val(uint8_t in) const {
		return invsbox[in];
	}

	inline uint8_t roundkey_byte(unsigned int round, unsigned int byte) const {
		return roundkeys[round*16 + byte];
	}
};


#endif /* AESHELPER_H_ */
