/*
 **
 ** Copyright 2019 Karlsruhe Institute of Technology, Chair of Dependable Nano Computing
 **
 ** Permission is hereby granted, free of charge, to any person obtaining a copy of this software
 ** and associated documentation files (the "Software"), to deal in the Software without restriction, 
 ** including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 ** and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do 
 ** so, subject to the following conditions:
 **
 ** The above copyright notice and this permission notice shall be included in all copies or substantial 
 ** portions of the Software.
 **
 ** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT 
 ** LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 ** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
 ** WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
 ** SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 **
 */

#ifndef CPA_H
#define CPA_H

#include <array>
#include <cinttypes>
#include <memory>
#include <cmath>
#include "AESHelper.h"
#include "types.h"

enum CPAMode {
    CPA_PLAIN,
    CPA_CIPHER
};

template <unsigned int tlen> class CPA
{
public:
    CPA(CPAMode m, uint8_t byte, uint8_t bit);
    void reset();
    void changeMode(CPAMode m);
    CPAMode getMode() const;
    void addTrace(const Sensordata &sense, const Block &aesblock);
    void getCurrentCorrelation(Correlation &corr) const;
    unsigned int getNumTraces() const;
private:
    std::array<uint64_t, 256> sumy;
    std::array<uint64_t, 256> sumy2;
    std::array<std::array<uint64_t, tlen>, 256> sumxy;
    std::array<uint64_t, tlen> sumx;
    std::array<uint64_t, tlen> sumx2;
    unsigned int n = 0;
    static const Block key;
    static const std::array<uint8_t, 16> hw_n;
    static inline uint8_t hw(uint8_t b) {
        return (hw_n[b&0x0f] + hw_n[b>>4]);
    }
    std::unique_ptr<AESHelper> aesfuncs;
    CPAMode m;
    uint8_t byte;
    uint8_t bit;
};

template <unsigned int tlen> const Block CPA<tlen>::key = {0x2b, 0x7e, 0x15, 0x16, 0x28, 0xae, 0xd2, 0xa6, 0xab, 0xf7, 0x15, 0x88, 0x09, 0xcf, 0x4f, 0x3c};
template <unsigned int tlen> const std::array<uint8_t, 16> CPA<tlen>::hw_n = {0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4};

template <unsigned int tlen> CPA<tlen>::CPA(CPAMode m, uint8_t byte, uint8_t bit) : m(m), byte(byte), bit(bit)
{
    this->aesfuncs = std::unique_ptr<AESHelper>(new AESHelper(key));
    this->reset();
}

template <unsigned int tlen> void CPA<tlen>::reset() {
    this->sumx.fill(0);
    this->sumx2.fill(0);
    this->sumy.fill(0);
    this->sumy2.fill(0);
    for (unsigned int keybyte = 0; keybyte < 256; keybyte++) {
        for (unsigned int t = 0; t < tlen; t++) {
            this->sumxy[keybyte][t] = 0;
        }
    }
    n = 0;
}

template <unsigned int tlen> void CPA<tlen>::changeMode(CPAMode m) {
    this->m = m;
}

template <unsigned int tlen> void CPA<tlen>::addTrace(const Sensordata &sense, const Block &aesblock) {
    std::array<uint8_t, 256> hypo;
    for (unsigned int keybyte = 0; keybyte < 256; keybyte++) { // Compute all power model values and update the respective sums
        if (this->m == CPAMode::CPA_PLAIN) {
            hypo[keybyte] = this->hw(this->aesfuncs->sbox_val(keybyte ^ aesblock[this->byte])&(0x01 << this->bit));
        } else {
            hypo[keybyte] = this->hw(this->aesfuncs->isbox_val(keybyte ^ aesblock[this->byte])&(0x01 << this->bit));
        }
        this->sumy[keybyte] += hypo[keybyte];
        this->sumy2[keybyte] += hypo[keybyte]*hypo[keybyte];
    }

    for (unsigned int i = 0; i < tlen; i++) { // Update sums with the values from the actual traces
        this->sumx[i] += sense[i];
        this->sumx2[i] += sense[i]*sense[i];
    }

    for (unsigned int keybyte = 0; keybyte < 256; keybyte++) { // Update the combined sum
        for (unsigned int t = 0; t < tlen; t++) {
            this->sumxy[keybyte][t] += hypo[keybyte]*sense[t];
        }
    }
    n++;
}

template <unsigned int tlen> void CPA<tlen>::getCurrentCorrelation(Correlation &corr) const {
    double n_d = static_cast<double>(n);
    for (unsigned int keybyte = 0; keybyte < 256; keybyte++) {
        corr[keybyte].resize(tlen);
        for (unsigned int i = 0; i < tlen; i++) {
            double denom = std::sqrt(sumx2[i] - sumx[i]*(sumx[i]/n_d))*std::sqrt(sumy2[keybyte] - sumy[keybyte]*(sumy[keybyte]/n_d));
            if (denom != 0.0) {
                corr[keybyte][i] = (sumxy[keybyte][i] - sumx[i]*(sumy[keybyte]/n_d))/denom;
            } else {
                corr[keybyte][i] = 0.0;
            }
        }
    }
}

template <unsigned int tlen> CPAMode CPA<tlen>::getMode() const {
    return this->m;
}

template <unsigned int tlen> unsigned int CPA<tlen>::getNumTraces() const {
    return this->n;
}

#endif // CPA_H
