/*
 **
 ** Copyright 2019 Karlsruhe Institute of Technology, Chair of Dependable Nano Computing
 **
 ** Permission is hereby granted, free of charge, to any person obtaining a copy of this software
 ** and associated documentation files (the "Software"), to deal in the Software without restriction, 
 ** including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 ** and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do 
 ** so, subject to the following conditions:
 **
 ** The above copyright notice and this permission notice shall be included in all copies or substantial 
 ** portions of the Software.
 **
 ** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT 
 ** LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 ** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
 ** WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
 ** SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 **
 */

#include "fpga.h"
#include <exception>
#include <thread>
#include <chrono>
#include <fcntl.h>
#include <stdio.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <chrono>

FPGA::FPGA(const std::string &serialport, unsigned int baudrate)
{
    this->serial_fd = open(serialport.c_str(), O_RDWR | O_NOCTTY);
    if (convRate(baudrate) == 0) {
        throw std::invalid_argument("Invalid baudrate.");
    }
    if (this->serial_fd < 0) {
        throw std::runtime_error("Could not open serial device.");
    }
    if (!isatty(this->serial_fd)) {
        close(this->serial_fd);
        throw std::runtime_error("Serial device is not a TTY device.");
    }
    struct termios config;
    if (tcgetattr(this->serial_fd, &config) < 0) {
        close(this->serial_fd);
        throw std::runtime_error("Could not get default TTY configuration.");
    }
    config.c_iflag = 0;
    config.c_oflag = 0;
    config.c_lflag = 0;
    config.c_cflag = CS8|CREAD|CLOCAL;
    config.c_cc[VMIN] = 0;
    config.c_cc[VTIME] = 10;
    if (cfsetspeed(&config, convRate(baudrate)) < 0) {
        close(this->serial_fd);
        throw std::runtime_error("Could not set the desired baudrate.");
    }
    tcsetattr(this->serial_fd, TCSANOW, &config);
}

FPGA::~FPGA() {
    if (this->serial_fd > 0) {
        close(this->serial_fd);
    }
}

void FPGA::setCycles(unsigned int ro_cycles) {
    this->ro_cycles = ro_cycles;
}

void FPGA::setActivecycles(unsigned int ro_activecycles) {
    this->ro_activecycles = ro_activecycles;
}

void FPGA::reset() {
    int status;
    if (ioctl(this->serial_fd, TIOCMGET, &status) < 0) {
        throw std::runtime_error("Could not use ioctl to get current TIOCM status.");
    }
    status |= TIOCM_RTS;
    if (ioctl(this->serial_fd, TIOCMSET, &status) < 0) {
        throw std::runtime_error("Could not use ioctl to set current TIOCM status.");
    }
    std::this_thread::sleep_for(std::chrono::seconds(1));
    if (ioctl(this->serial_fd, TIOCMGET, &status) < 0) {
        throw std::runtime_error("Could not use ioctl to get current TIOCM status.");
    }
    status &= ~TIOCM_RTS;
    if (ioctl(this->serial_fd, TIOCMSET, &status) < 0) {
        throw std::runtime_error("Could not use ioctl to set current TIOCM status.");
    }
}

void FPGA::resetBlock() {
    int status;
    if (ioctl(this->serial_fd, TIOCMGET, &status) < 0) {
        throw std::runtime_error("Could not use ioctl to get current TIOCM status.");
    }
    status |= TIOCM_RTS;
    if (ioctl(this->serial_fd, TIOCMSET, &status) < 0) {
        throw std::runtime_error("Could not use ioctl to set current TIOCM status.");
    }
}

void FPGA::release() {
    int status;
    if (ioctl(this->serial_fd, TIOCMGET, &status) < 0) {
        throw std::runtime_error("Could not use ioctl to get current TIOCM status.");
    }
    status &= ~TIOCM_RTS;
    if (ioctl(this->serial_fd, TIOCMSET, &status) < 0) {
        throw std::runtime_error("Could not use ioctl to set current TIOCM status.");
    }
}

bool FPGA::inReset() const {
    int status;
    if (ioctl(this->serial_fd, TIOCMGET, &status) < 0) {
        throw std::runtime_error("Could not use ioctl to get current TIOCM status.");
    }
    return (status & TIOCM_RTS);
}

void FPGA::setROParameters(unsigned int ro_activecycles, unsigned int ro_cycles) {
    this->ro_cycles = ro_cycles;
    this->ro_activecycles = ro_activecycles;
}

unsigned int FPGA::getCycles() const {
    return this->ro_cycles;
}

unsigned int FPGA::getActivecycles() const {
    return this->ro_activecycles;
}

speed_t FPGA::convRate(unsigned int baudrate) const {
    switch (baudrate) {
        case 9600:
            return B9600;
        case 19200:
            return B19200;
        case 38400:
            return B38400;
        case 57600:
            return B57600;
        case 115200:
            return B115200;
        case 230400:
            return B230400;
        case 460800:
            return B460800;
        case 500000:
            return B500000;
        case 576000:
            return B576000;
        case 921600:
            return B921600;
        case 1000000:
            return B1000000;
        case 1152000:
            return B1152000;
        case 1500000:
            return B1500000;
        case 2000000:
            return B2000000;
        case 2500000:
            return B2500000;
        case 3000000:
            return B3000000;
        case 3500000:
            return B3500000;
        case 4000000:
            return B4000000;
        default:
            return 0;
    }
}

void FPGA::crash() {
    uint8_t s = 0x73;
    write(this->serial_fd, &s, 1);
}

bool FPGA::encryptGeneric(const Block &plaintext, Block &cipher, uint8_t ro_ena) {
    if (this->serial_fd < 0) {
        throw std::runtime_error("UART has not been initialized.");
    }
    uint8_t buffer_activecycles[4];
    uint8_t buffer_cycles[4];
    for (unsigned int i = 0; i < 4; i++) {
        buffer_cycles[i] = 0xff & ((this->ro_cycles) >> (24 - i*8));
        buffer_activecycles[i] = 0xff & ((this->ro_activecycles) >> (24 - i*8));
    }
    write(this->serial_fd, &ro_ena, 1);
    write(this->serial_fd, buffer_cycles, 4);
    write(this->serial_fd, buffer_activecycles, 4);
    write(this->serial_fd, plaintext.data(), 16);

    auto start = std::chrono::steady_clock::now();
    ssize_t len = read(this->serial_fd, cipher.data(), 16);
    while (len < 16) {
        len += read(this->serial_fd, cipher.data()+len, 16);
        auto elapsed = std::chrono::duration_cast<std::chrono::milliseconds>(std::chrono::steady_clock::now() - start);
        if (elapsed.count() > UART_TIMEOUT) {
            this->reset();
            return false;
        }
    }
    return true;
}

bool FPGA::encryptSCA(const Block &plaintext, Block &cipher, Sensordata &sensordata) {
    if (this->serial_fd < 0) {
        throw std::runtime_error("UART has not been initialized.");
    }

    write(this->serial_fd, plaintext.data(), 16);
    auto start = std::chrono::steady_clock::now();
    ssize_t len = read(this->serial_fd, cipher.data(), 16);
    while (len < 16) {
        len += read(this->serial_fd, cipher.data()+len, static_cast<unsigned int>(16-len));
        auto elapsed = std::chrono::duration_cast<std::chrono::milliseconds>(std::chrono::steady_clock::now() - start);
        if (elapsed.count() > UART_TIMEOUT) {
            this->reset();
            return false;
        }
    }
    sensordata.resize(this->sca_tracelen);
    start = std::chrono::steady_clock::now();
    len = read(this->serial_fd, sensordata.data(), this->sca_tracelen);
    while (len < this->sca_tracelen) {
        len += read(this->serial_fd, sensordata.data() + len, this->sca_tracelen);
        auto elapsed = std::chrono::duration_cast<std::chrono::milliseconds>(std::chrono::steady_clock::now() - start);
        if (elapsed.count() > UART_TIMEOUT) {
            this->reset();
            return false;
        }
    }
    return true;
}

bool FPGA::encrypt(const Block &plaintext, Block &cipher) {
    return this->encryptGeneric(plaintext, cipher, 0x00);
}

bool FPGA::encryptFaulty(const Block &plaintext, Block &cipher) {
    return this->encryptGeneric(plaintext, cipher, 0x01);
}

