/*
 **
 ** Copyright 2019 Karlsruhe Institute of Technology, Chair of Dependable Nano Computing
 **
 ** Permission is hereby granted, free of charge, to any person obtaining a copy of this software
 ** and associated documentation files (the "Software"), to deal in the Software without restriction, 
 ** including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 ** and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do 
 ** so, subject to the following conditions:
 **
 ** The above copyright notice and this permission notice shall be included in all copies or substantial 
 ** portions of the Software.
 **
 ** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT 
 ** LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 ** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
 ** WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
 ** SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 **
 */

#include "mainwindow.h"
#include "ui_mainwindow.h"
#include <QMessageBox>
#include <QtCharts/QLineSeries>
#include <sstream>
#include <QTextStream>
#include <iomanip>
#include <QtCharts/QValueAxis>
#include <stdlib.h>

const Block MainWindow::key = {0x2b, 0x7e, 0x15, 0x16, 0x28, 0xae, 0xd2, 0xa6, 0xab, 0xf7, 0x15, 0x88, 0x09, 0xcf, 0x4f, 0x3c};
const Block MainWindow::roundkey = {0xd0, 0x14, 0xf9, 0xa8, 0xc9, 0xee, 0x25, 0x89, 0xe1, 0x3f, 0x0c, 0xc8, 0xb6, 0x63, 0x0c, 0xa6};

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow)
{
    qRegisterMetaType<Block>();
    qRegisterMetaType<Sensordata>();
    qRegisterMetaType<Correlation>();
    qRegisterMetaType<CorrelationProgress>();
    ui->setupUi(this);
    this->ui->modeSel->addItem(QString("Plaintext based"));
    this->ui->modeSel->addItem(QString("Ciphertext based"));
    this->ui->modeSel->setCurrentIndex(1);
    this->ui->sca_total->setChart(new QtCharts::QChart());
    this->ui->sca_progress->setChart(new QtCharts::QChart());
}

MainWindow::~MainWindow()
{
    if (dfathread && dfathread->isRunning()) {
        dfathread->setQuit();
        dfathread->wait();
    }
    if (scathread && scathread->isRunning()) {
        scathread->setQuit();
        scathread->wait();
    }
    delete ui;
}

std::string MainWindow::byteToString(uint8_t a) const {
    std::stringstream ss;
    ss << std::setfill('0') << std::setw(2) << std::hex << static_cast<int>(a);
    return ss.str();
}

QString MainWindow::arrayToString(const Block &a) const {
    QString as;
    for (unsigned int i = 0; i < 15; i++) {
        as += QString(this->byteToString(a[i]).c_str());
        as += QString(":");
    }
    as += QString(this->byteToString(a[15]).c_str());
    return as;
}

QString MainWindow::faultyArrayToString(const Block &correct, const Block &faulty) const {
    QString as;
    for (unsigned int i = 0; i < 16; i++) {
        if (correct[i] != faulty[i]) {
            as += "<span style='color: #f00'>";
            as += QString(this->byteToString(faulty[i]).c_str());
            as += "</span>";
        } else {
            as += "<span style='color: #0f0'>";
            as += QString(this->byteToString(faulty[i]).c_str());
            as += "</span>";
        }
        if (i < 15)
            as += QString("<span style='color: #000'>:</span>");
    }
    return as;
}

void MainWindow::on_toggleDFAUART_clicked()
{
    if (dfaFPGA) {
        this->dfathread->setQuit();
        this->dfathread->wait();
        this->dfathread = nullptr;
        this->dfaFPGA = nullptr;
        this->ui->roFreq->setEnabled(false);
        this->ui->roDutyCycle->setEnabled(false);
        this->ui->toggleEnc->setEnabled(false);
        this->ui->toggleFault->setEnabled(false);
        this->ui->toggleShowEnc->setEnabled(false);
        this->ui->dfafpgaReset->setEnabled(false);
        this->ui->toggleDFAUART->setText("Open");
        this->ui->toggleEnc->setText(QString("Start Encryptions"));
        this->ui->toggleFault->setText(QString("Activate Injection"));
        this->showEnc = false;
        this->ui->toggleShowEnc->setText(QString("Show Encryptions"));
    }   else {
        try {
            this->dfaFPGA = std::shared_ptr<FPGA>(new FPGA(this->ui->dfa_port->text().toStdString(), this->ui->dfa_baudrate->text().toUInt()));
            this->dfathread = std::shared_ptr<DFAThread> (new DFAThread(this, dfaFPGA));
            connect(this->dfathread.get(), &DFAThread::encryptionFinished, this, &MainWindow::encryptionFinishedDFA);
            connect(this->dfathread.get(), &DFAThread::faultInjected, this, &MainWindow::faultInjected);
            this->dfathread->start();
            unsigned int ro_cycles = this->FPGA_FREQ/this->ui->roFreq->text().toUInt();
            this->dfaFPGA->setCycles(ro_cycles);
            this->ui->roFreq->setText(QString::number(this->FPGA_FREQ/ro_cycles));
            unsigned int ro_activecycles = static_cast<unsigned int>(this->dfaFPGA->getCycles()*this->ui->roDutyCycle->text().toDouble()/100.0);
            this->dfaFPGA->setActivecycles(ro_activecycles);
            this->ui->roDutyCycle->setText(QString::number(static_cast<double>(ro_activecycles)*100.0/this->dfaFPGA->getCycles(), 'f', 2));
            this->ui->toggleDFAUART->setText("Close");
            this->dfaFPGA->reset();
            this->ui->roFreq->setEnabled(true);
            this->ui->roDutyCycle->setEnabled(true);
            this->ui->toggleEnc->setEnabled(true);
            this->ui->toggleFault->setEnabled(true);
            this->ui->toggleShowEnc->setEnabled(true);
            this->ui->dfafpgaReset->setEnabled(true);
            this->ui->toggleEnc->setText(QString("Start Encryptions"));
            this->ui->toggleFault->setText(QString("Activate Injection"));
            this->showEnc = false;
            this->ui->toggleShowEnc->setText(QString("Show Encryptions"));
        } catch (std::exception &e) {
            QMessageBox::critical(this, QString("Error"), QString("Error opening the UART device: ") + e.what(), QMessageBox::Ok);
            this->ui->toggleDFAUART->setText("Open");
            dfathread = nullptr;
            dfaFPGA = nullptr;
            this->ui->roFreq->setEnabled(false);
            this->ui->roDutyCycle->setEnabled(false);
            this->ui->toggleEnc->setEnabled(false);
            this->ui->toggleFault->setEnabled(false);
            this->ui->toggleShowEnc->setEnabled(false);
            this->ui->dfafpgaReset->setEnabled(false);
            this->ui->toggleEnc->setText(QString("Start Encryptions"));
            this->ui->toggleFault->setText(QString("Activate Injection"));
            this->showEnc = false;
            this->ui->toggleShowEnc->setText(QString("Show Encryptions"));
        }
    }
}

void MainWindow::on_roFreq_editingFinished()
{
    unsigned int ro_cycles = this->FPGA_FREQ/this->ui->roFreq->text().toUInt();
    unsigned int ro_activecycles = 0;
    if (dfaFPGA) {
        dfaFPGA->setCycles(ro_cycles);
        ro_activecycles = static_cast<unsigned int>(this->dfaFPGA->getCycles()*this->ui->roDutyCycle->text().toDouble()/100.0);
        dfaFPGA->setActivecycles(ro_activecycles);
    }
    this->ui->roFreq->setText(QString::number(this->FPGA_FREQ/ro_cycles));
    this->ui->roDutyCycle->setText(QString::number((static_cast<double>(ro_activecycles)*100.0/this->dfaFPGA->getCycles())));
}

void MainWindow::on_roDutyCycle_editingFinished()
{
    unsigned int ro_activecycles = 0;
    if (dfaFPGA) {
        ro_activecycles = static_cast<unsigned int>(this->dfaFPGA->getCycles()*this->ui->roDutyCycle->text().toDouble()/100.0);
        dfaFPGA->setActivecycles(ro_activecycles);
    }
    this->ui->roDutyCycle->setText(QString::number((static_cast<double>(ro_activecycles)*100.0/this->dfaFPGA->getCycles())));
}

void MainWindow::encryptionFinishedDFA(const Block &plaintext, const Block &cipher) {
    if (this->showEnc) {
        this->ui->enc_text->appendHtml(QString("<span style='color: #000'>Encrypted: </span>"));
        this->ui->enc_text->appendHtml(QString("<span style='color: #000'>Plaintext:  ") + this->arrayToString(plaintext) + QString("</span>"));
        this->ui->enc_text->appendHtml(QString("<span style='color: #000'>Ciphertext: ") + this->arrayToString(cipher) + QString("</span>\n"));
    }
}

void MainWindow::encryptionFinishedSCA(const Block &plaintext, const Block &cipher, const Sensordata &sensordata) {
    this->ui->num_traces->setText(QString::number(this->scathread->getNumTraces()));
}

void MainWindow::correlationUpdated(const Correlation &corr, const CorrelationProgress &prog) {
    QtCharts::QChart *chart = this->ui->sca_total->chart();
    chart->removeAllSeries();
    QtCharts::QLineSeries *correct = nullptr;
    for (unsigned int keybyte = 0; keybyte < 256; keybyte++) {
        QtCharts::QLineSeries *series = new QtCharts::QLineSeries();
        for (unsigned int t = 0; t < corr.size(); t++) {
            series->append(t, corr[keybyte][t]);
        }
        if (((this->scathread->getCurrentMode() == CPAMode::CPA_PLAIN) && (keybyte == this->key[this->scathread->getCurrentByte()]))
                || ((this->scathread->getCurrentMode() == CPAMode::CPA_CIPHER) && (keybyte == this->roundkey[this->scathread->getCurrentByte()]))) {
            series->setColor(QColor(255,0,0));
            correct = series;
        } else {

            series->setColor(QColor(200,200,200));
            chart->addSeries(series);
        }
    }
    chart->legend()->setVisible(false);
    chart->addSeries(correct);
    chart->createDefaultAxes();
    QtCharts::QValueAxis *xAxis = new QtCharts::QValueAxis();
    xAxis->setMin(0);
    xAxis->setMax(corr[0].size());
    xAxis->setTickCount(10);
    xAxis->setLabelFormat("%d");
    chart->setAxisX(xAxis);

    chart = this->ui->sca_progress->chart();
    chart->removeAllSeries();
    correct = nullptr;
    for (unsigned int keybyte = 0; keybyte < 256; keybyte++) {
        QtCharts::QLineSeries *series = new QtCharts::QLineSeries();
        unsigned int pstep = this->scathread->getProgressStep();
        unsigned int step = 0;
        for (auto val : prog[keybyte]) {
            series->append(step, val);
            step += pstep;
        }
        if (((this->scathread->getCurrentMode() == CPAMode::CPA_PLAIN) && (keybyte == this->key[this->scathread->getCurrentByte()]))
                || ((this->scathread->getCurrentMode() == CPAMode::CPA_CIPHER) && (keybyte == this->roundkey[this->scathread->getCurrentByte()]))) {
            series->setColor(QColor(255,0,0));
            correct = series;
        } else {
            series->setColor(QColor(200,200,200));
            chart->addSeries(series);
        }
    }
    chart->legend()->setVisible(false);
    chart->addSeries(correct);
    chart->createDefaultAxes();
    xAxis = new QtCharts::QValueAxis();
    xAxis->setMin(0);
    xAxis->setMax(prog[0].size()*this->scathread->getProgressStep());
    xAxis->setTickCount(10);
    xAxis->setLabelFormat("%d");
    chart->setAxisX(xAxis);
}

void MainWindow::faultInjected(const Block &correct_cipher, const Block &faulty_cipher) {
    if (this->cracker.isFeasibile(correct_cipher, faulty_cipher)) {
        this->ui->enc_text->appendHtml(QString("<span style='color: #f00'>Feasible fault injected:</span>"));
        unsigned int type = static_cast<unsigned int>(this->cracker.determineFaultType(correct_cipher, faulty_cipher));
        this->num_faults_per_cat[type]++;
        this->ui->faultsCat0->setText(QString::number(this->num_faults_per_cat[0]));
        this->ui->faultsCat1->setText(QString::number(this->num_faults_per_cat[1]));
        this->ui->faultsCat2->setText(QString::number(this->num_faults_per_cat[2]));
        this->ui->faultsCat3->setText(QString::number(this->num_faults_per_cat[3]));
        this->cracker.evaluatePair(correct_cipher, faulty_cipher);
        unsigned int log2_cand = this->cracker.getTotalLog2Candidates();
        if (log2_cand == 0) {
            this->ui->keyFoundAlert->setText(QString("Key found!"));
            this->dfathread->setEnc(false);
            this->ui->toggleEnc->setText(QString("Start Encryptions"));
            this->dfathread->setFault(false);
            this->ui->toggleFault->setText(QString("Activate Injection"));
            this->showEnc = false;
            this->ui->toggleShowEnc->setText(QString("Show Encryptions"));
            QMessageBox::information(this, QString("Key found!"), QString("Correct key: ") + this->arrayToString(this->cracker.recoverAESKey(this->cracker.getRoundKeyCandidate())));
        }
        this->ui->keyCandidates->setText(QString("2^") + QString::number(log2_cand));
    } else {
        this->ui->enc_text->appendHtml(QString("<span style='color: #f00'>Fault injected:</span>"));
    }
    this->ui->enc_text->appendHtml(QString("<span style='color: #000'>Correct ciphertext: ") + this->arrayToString(correct_cipher) + QString("</span>"));
    this->ui->enc_text->appendHtml(QString("<span style='color: #000'>Faulty ciphertext:  </span>") + this->faultyArrayToString(correct_cipher, faulty_cipher) + QString("\n"));
}

void MainWindow::on_toggleEnc_clicked()
{
    if (this->dfathread->getEnc()) {
        this->ui->toggleEnc->setText(QString("Start Encryptions"));
    } else {    
        this->ui->toggleEnc->setText(QString("Stop Encryptions"));
    }
    this->dfathread->toggleEnc();
}

void MainWindow::on_toggleFault_clicked()
{
    if (this->dfathread->getFault()) {
        this->ui->toggleFault->setText(QString("Activate Injection"));
    } else {
        this->ui->toggleFault->setText(QString("Deactivate Injection"));
    }
    this->dfathread->toggleFault();
}

void MainWindow::on_toggleShowEnc_clicked()
{
    if (this->showEnc) {
        this->showEnc = false;
        this->ui->toggleShowEnc->setText(QString("Show Encryptions"));
    } else {
        this->showEnc = true;
        this->ui->toggleShowEnc->setText(QString("Hide Encryptions"));
    }
}

void MainWindow::on_resetRecovery_clicked()
{
    this->cracker.reset();
    this->num_faults_per_cat = {{0, 0, 0, 0}};
    this->ui->faultsCat0->setText(QString::number(this->num_faults_per_cat[0]));
    this->ui->faultsCat1->setText(QString::number(this->num_faults_per_cat[1]));
    this->ui->faultsCat2->setText(QString::number(this->num_faults_per_cat[2]));
    this->ui->faultsCat3->setText(QString::number(this->num_faults_per_cat[3]));
    this->ui->keyFoundAlert->setText(QString(""));
    this->ui->keyCandidates->setText(QString("2^128"));
}

void MainWindow::on_dfafpgaReset_clicked()
{
    if (this->dfaFPGA) {
        this->dfathread->setQuit();
        this->dfathread->wait();
        this->dfathread = nullptr;
        this->ui->toggleFault->setEnabled(false);
        this->ui->toggleEnc->setText(QString("Start Encryptions"));
        this->ui->toggleEnc->setEnabled(false);
        this->ui->toggleFault->setText(QString("Activate Injection"));
        this->ui->toggleFault->setEnabled(false);
        this->showEnc = false;
        this->ui->toggleShowEnc->setText(QString("Show Encryptions"));
        this->dfaFPGA->reset();
        this->dfathread = std::shared_ptr<DFAThread> (new DFAThread(this, dfaFPGA));
        connect(this->dfathread.get(), &DFAThread::encryptionFinished, this, &MainWindow::encryptionFinishedDFA);
        connect(this->dfathread.get(), &DFAThread::faultInjected, this, &MainWindow::faultInjected);
        this->dfathread->start();
        this->ui->toggleEnc->setEnabled(true);
        this->ui->toggleFault->setEnabled(true);
        this->ui->toggleFault->setEnabled(true);
    }
}

void MainWindow::on_toggleSCAUART_clicked()
{
    if (scaFPGA) {
        this->scathread->setQuit();
        this->scathread->wait();
        this->scathread = nullptr;
        this->scaFPGA = nullptr;
        this->ui->scafpgaReset->setEnabled(false);
        this->ui->scaToggleEnc->setEnabled(false);
        this->ui->byteSel->setEnabled(false);
        this->ui->bitSel->setEnabled(false);
        this->ui->pstep->setEnabled(false);
        this->ui->ppoint->setEnabled(false);
        this->ui->toggleSCAUART->setText("Open");
        this->ui->scaToggleEnc->setText(QString("Start Encryptions"));
    }   else {
        try {
            this->scaFPGA = std::shared_ptr<FPGA>(new FPGA(this->ui->sca_port->text().toStdString(), this->ui->sca_baudrate->text().toUInt()));
            this->scathread = std::shared_ptr<SCAThread> (new SCAThread(this, scaFPGA));
            connect(this->scathread.get(), &SCAThread::encryptionFinished, this, &MainWindow::encryptionFinishedSCA);
            connect(this->scathread.get(), &SCAThread::correlationUpdated, this, &MainWindow::correlationUpdated);
            this->scathread->start();
            this->ui->toggleSCAUART->setText("Close");
            this->scaFPGA->reset();
            this->ui->scaToggleEnc->setEnabled(true);
            this->ui->scafpgaReset->setEnabled(true);
            this->ui->scaToggleEnc->setText(QString("Start Encryptions"));
            this->scathread->setCurrentByte(static_cast<unsigned int>(this->ui->byteSel->value()));
            this->scathread->setCurrentBit(static_cast<unsigned int>(this->ui->bitSel->value()));
            int ppoint = this->ui->ppoint->text().toInt();
            this->scathread->setProgressPoint(ppoint);
            unsigned int pstep = this->ui->pstep->text().toUInt();
            this->scathread->setProgressStep(pstep);
            this->ui->byteSel->setEnabled(true);
            this->ui->bitSel->setEnabled(true);
            this->ui->pstep->setEnabled(true);
            this->ui->ppoint->setEnabled(true);
        } catch (std::exception &e) {
            QMessageBox::critical(this, QString("Error"), QString("Error opening the UART device: ") + e.what(), QMessageBox::Ok);
            this->ui->toggleDFAUART->setText("Open");
            this->scathread = nullptr;
            this->scaFPGA = nullptr;
            this->ui->scafpgaReset->setEnabled(false);
            this->ui->scaToggleEnc->setEnabled(false);
            this->ui->toggleSCAUART->setText("Open");
            this->ui->scaToggleEnc->setText(QString("Start Encryptions"));
            this->ui->byteSel->setEnabled(false);
            this->ui->bitSel->setEnabled(false);
            this->ui->pstep->setEnabled(false);
            this->ui->ppoint->setEnabled(false);
        }
    }
}

void MainWindow::on_scafpgaReset_clicked()
{
    if (this->scaFPGA) {
        this->scathread->setQuit();
        this->scathread->wait();
        this->scathread = nullptr;
        this->ui->scafpgaReset->setEnabled(false);
        this->ui->scaToggleEnc->setEnabled(false);
        this->ui->toggleEnc->setText(QString("Start Encryptions"));
        this->scaFPGA->reset();
        this->scathread = std::shared_ptr<SCAThread> (new SCAThread(this, scaFPGA));
        connect(this->scathread.get(), &SCAThread::encryptionFinished, this, &MainWindow::encryptionFinishedSCA);
        this->scathread->start();
        this->ui->scaToggleEnc->setEnabled(true);
        this->ui->scafpgaReset->setEnabled(true);
    }
}

void MainWindow::on_scaToggleEnc_clicked()
{
    if (this->scathread->getEnc()) {
        this->ui->scaToggleEnc->setText(QString("Start Encryptions"));
    } else {
        this->ui->scaToggleEnc->setText(QString("Stop Encryptions"));
    }
    this->scathread->toggleEnc();
}

void MainWindow::on_resetCPA_clicked()
{
    this->scathread->resetCPAs();
}

void MainWindow::on_ppoint_editingFinished()
{
    int ppoint = this->ui->ppoint->text().toInt();
    this->scathread->setProgressPoint(ppoint);
}

void MainWindow::on_pstep_editingFinished()
{
    unsigned int pstep = this->ui->pstep->text().toUInt();
    this->scathread->setProgressStep(pstep);
}

void MainWindow::on_byteSel_valueChanged(int arg1)
{
    this->ui->sca_total->chart()->removeAllSeries();
    this->ui->sca_progress->chart()->removeAllSeries();
    this->scathread->setCurrentByte(static_cast<unsigned int>(arg1));
}

void MainWindow::on_bitSel_valueChanged(int arg1)
{
    this->ui->sca_total->chart()->removeAllSeries();
    this->ui->sca_progress->chart()->removeAllSeries();
    this->scathread->setCurrentBit(static_cast<unsigned int>(arg1));
}

void MainWindow::on_modeSel_currentIndexChanged(int index)
{
    if (this->scathread) {
        switch (index) {
            case 0: this->scathread->setCurrentMode(CPAMode::CPA_PLAIN);
                    break;
            case 1: this->scathread->setCurrentMode(CPAMode::CPA_CIPHER);
                    break;
            default: this->scathread->setCurrentMode(CPAMode::CPA_CIPHER);
        }
    }
}

void MainWindow::on_toggleCrashUART_clicked()
{
    if (crashFPGA) {
        this->crashFPGA = nullptr;
        this->ui->crashfpgaReset->setEnabled(false);
        this->ui->toggleCrashUART->setText("Open");
        this->ui->startCrash->setEnabled(false);
    }   else {
        try {
            this->crashFPGA = std::shared_ptr<FPGA>(new FPGA(this->ui->crash_port->text().toStdString(), this->ui->crash_baudrate->text().toUInt()));
            this->ui->toggleCrashUART->setText("Close");
            this->crashFPGA->reset();
            this->ui->crashfpgaReset->setEnabled(true);
            this->ui->startCrash->setEnabled(true);
        } catch (std::exception &e) {
            QMessageBox::critical(this, QString("Error"), QString("Error opening the UART device: ") + e.what(), QMessageBox::Ok);
            this->crashFPGA = nullptr;
            this->ui->crashfpgaReset->setEnabled(false);
            this->ui->startCrash->setEnabled(false);
            this->ui->toggleCrashUART->setText("Open");
        }
    }
}

void MainWindow::on_crashfpgaReset_clicked()
{
    if (this->crashFPGA) {
        this->crashFPGA->reset();
    }
}

void MainWindow::on_startCrash_clicked()
{
    if (this->crashFPGA) {
        this->crashFPGA->crash();
    }
}

void MainWindow::on_dosReprogram_clicked()
{
    std::stringstream ss;
    ss << "iceprog -S ./dos.bin -d d:" << this->ui->programDevid->text().toStdString();
    system(ss.str().c_str());
}
