/*
 **
 ** Copyright 2019 Karlsruhe Institute of Technology, Chair of Dependable Nano Computing
 **
 ** Permission is hereby granted, free of charge, to any person obtaining a copy of this software
 ** and associated documentation files (the "Software"), to deal in the Software without restriction, 
 ** including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 ** and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do 
 ** so, subject to the following conditions:
 **
 ** The above copyright notice and this permission notice shall be included in all copies or substantial 
 ** portions of the Software.
 **
 ** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT 
 ** LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
 ** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
 ** WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
 ** SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 **
 */

#include "scathread.h"

SCAThread::SCAThread(QObject *parent, const std::shared_ptr<FPGA> &scafpga) : QThread(parent), scafpga(scafpga)
{
    *(this->encEna) = false;
    *(this->quitThread) = false;
    *(this->reset) = false;
    *(this->current_bit) = 0;
    *(this->current_byte) = 0;
    *(this->current_mode) = CPAMode::CPA_CIPHER;
    *(this->pstep) = 100;
    *(this->ppoint) = -1;
    for (unsigned int byte = 0; byte < 16; byte++) {
        for (unsigned int bit = 0; bit < 8; bit++) {
            this->cpas[byte][bit] = std::shared_ptr<CPA<TRACELEN>>(new CPA<TRACELEN>(CPAMode::CPA_CIPHER, static_cast<uint8_t>(byte), static_cast<uint8_t>(bit)));
        }
    }
}

void SCAThread::fillRandom(Block &target) {
    for (unsigned int i = 0; i < 16; i++) {
        target[i] = this->dist(this->gen);
    }
}

void SCAThread::run() {
    while(!*(this->quitThread)) {
        if(*(this->encEna)) {
            Block plaintext;
            Block cipher;
            Sensordata sensordata;
            this->fillRandom(plaintext);
            if (!this->scafpga->encryptSCA(plaintext, cipher, sensordata)) {
                continue;
            }
            emit encryptionFinished(plaintext, cipher, sensordata);
            for (unsigned int byte = 0; byte < 16; byte++) {
                for (unsigned int bit = 0; bit < 8; bit++) {
                    if (this->cpas[byte][bit]->getMode() != *(this->current_mode)) {
                        this->cpas[byte][bit]->reset();
                        this->cpas[byte][bit]->changeMode(*(this->current_mode));
                    }
                    if (this->cpas[byte][bit]->getMode() == CPAMode::CPA_PLAIN) {
                        this->cpas[byte][bit]->addTrace(sensordata, plaintext);
                    } else {
                        this->cpas[byte][bit]->addTrace(sensordata, cipher);
                    }
                    if ((this->cpas[byte][bit]->getNumTraces() > 0) && (this->cpas[byte][bit]->getNumTraces() % *(this->pstep) == 0)) {
                        Correlation corr;
                        this->cpas[byte][bit]->getCurrentCorrelation(corr);
                        int ppoint = *(this->ppoint);
                        if (ppoint < 0) {
                            double maxcorr = 0.0;
                            unsigned int maxpoint = 0;
                            for (unsigned int keybyte = 0; keybyte < 256; keybyte++) {
                                for (unsigned int t = 0; t < corr[0].size(); t++) {
                                    if (std::abs(corr[keybyte][t]) > maxcorr) {
                                        maxpoint = t;
                                        maxcorr = std::abs(corr[keybyte][t]);
                                    }
                                }
                            }
                            ppoint = static_cast<int>(maxpoint);
                        }
                        for (unsigned int keybyte = 0; keybyte < 256; keybyte++) {
                            progress[byte][bit][keybyte].push_back(corr[keybyte][static_cast<unsigned int>(ppoint)]);
                            if (progress[byte][bit][keybyte].size() > PROGRESS_HIST) {
                                progress[byte][bit][keybyte].pop_front();
                            }
                        }
                        if ((byte == *(this->current_byte)) && (bit == *(this->current_bit))) {
                            emit correlationUpdated(corr, progress[byte][bit]);
                        }
                    }
                }
            }
        }
        if (*(this->reset)) {
            for (unsigned int byte = 0; byte < 16; byte++) {
                for (unsigned int bit = 0; bit < 8; bit++) {
                    this->cpas[byte][bit]->reset();
                }
            }
            *(this->reset) = false;
        }
    }
    this->quit();
}

void SCAThread::resetCPAs() {
    *(this->reset) = true;
}

unsigned int SCAThread::getNumTraces() const {
    return this->cpas[*(this->current_byte)][*(this->current_bit)]->getNumTraces();
}

void SCAThread::setCurrentBit(unsigned int bit) {
    *(this->current_bit) = bit;
}

void SCAThread::setCurrentByte(unsigned int byte) {
    *(this->current_byte) = byte;
}

void SCAThread::setProgressStep(unsigned int pstep) {
    *(this->pstep) = pstep;
}

void SCAThread::setProgressPoint(int ppoint) {
    *(this->ppoint) = ppoint;
}

void SCAThread::setCurrentMode(CPAMode m) {
    *(this->current_mode) = m;
}

void SCAThread::toggleEnc() {
    *(this->encEna) = !*(this->encEna);
}

void SCAThread::setEnc(bool status) {
    *(this->encEna) = status;
}

bool SCAThread::getEnc() const {
    return *(this->encEna);
}

void SCAThread::setQuit() {
    *(this->quitThread) = true;
}

CPAMode SCAThread::getCurrentMode() const {
    return this->cpas[*(this->current_byte)][*(this->current_bit)]->getMode();
}

unsigned int SCAThread::getCurrentByte() const {
    return *(this->current_byte);
}

unsigned int SCAThread::getProgressStep() const {
    return *(this->pstep);
}
