----------------------------------------------------------------------------------
--
-- Copyright 2019 Ruhr University Bochum, Horst Görtz Institute for IT Security
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy of this software
-- and associated documentation files (the "Software"), to deal in the Software without restriction, 
-- including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, 
-- and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do 
-- so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in all copies or substantial 
-- portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT 
-- LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
-- IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
-- WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
-- SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
--
----------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

entity nxmBitShiftReg is
	generic (
		N		: integer range 2 to 256 := 4;					-- number of registers, IMPORTANT: N = 1 is not allowed (multiplexer config)
		M 		: integer range 1 to 256 := 2					-- bits per register
	);
	port (
		CLK		: in std_logic;									-- input clock
		SR		: in std_logic;									-- set/reset signal
		SRINIT	: in std_logic_vector(n*m-1 downto 0);			-- init value for reset
		CE		: in std_logic;									-- enable signal
		OPMODE	: in std_logic_vector(1 downto 0);				-- operation mode (see list of modes)
		DIN		: in std_logic_vector(m-1 downto 0);			-- input value when shifting
		DOUT	: out std_logic_vector(m-1 downto 0);			-- output: word shifted out resp. last word rotated
		DOUT_F	: out std_logic_vector(n*m-1 downto 0)			-- output: complete register
	);
end nxmBitShiftReg;

architecture Behavioral of nxmBitShiftReg is
	-- --------------------------------------------- 
	--                Components
	-- --------------------------------------------- 

	-- synchronously reset n bit register
	component nBitReg_SSR is
		generic (
			BIT_WIDTH	: integer range 1 to 256 := 16				-- register size in bits
		);
		port (
			CLK		: in std_logic;									-- input clock
			SR		: in std_logic;									-- set/reset (high)
			SRINIT	: in std_logic_vector(BIT_WIDTH-1 downto 0);	-- init value
			CE		: in std_logic;									-- enable signal
			DIN		: in std_logic_vector(BIT_WIDTH-1 downto 0);	-- input word
			DOUT	: out std_logic_vector(BIT_WIDTH-1 downto 0)	-- output word
		);
	end component nBitReg_SSR;

	-- --------------------------------------------- 
	--                Signals
	-- --------------------------------------------- 
	signal opmode_intern	: std_logic_vector(1 downto 0);
		alias rotate		: std_logic is opmode_intern(1);
		alias toleft		: std_logic is opmode_intern(0);
	
	signal din_intern : std_logic_vector(n*m-1 downto 0);
	signal dout_intern : std_logic_vector(n*m-1 downto 0);
begin

	-- assign OPMODE to internal signal in order to use pre-defined alias
	opmode_intern <= OPMODE;

	-- generate shift register
	shiftReg : for i in 1 to n generate
	
		-- registers for the shift register
		reg : nBitReg_SSR generic map ( BIT_WIDTH => M ) 
				port map (
					CLK 	=> CLK,
					SR		=> SR,
					SRINIT	=> SRINIT(i*m-1 downto (i-1)*m),
					CE		=> CE,
					DIN		=> din_intern(i*m-1 downto (i-1)*m),
					DOUT	=> dout_intern(i*m-1 downto (i-1)*m)
				);
		-- ---------------------
		
		-- wire input and output
		-- ---------------------
		
		-- input signal multiplexer configuration concerning DIN and output index
		-- --------------------------------------
		-- left	rot		i = n		i		i = 1
		-- --------------------------------------
		--	0	0		DIN			i+1		i+1
		--	0	1		i=1			i+1		i+1
		--	1	0		i-1			i-1		DIN
		--	1	1		i-1			i-1		i=n

		
		-- left most: choose between DIN, out(1), out(i-1)
		in_lmr: if i = n generate
			din_intern(i*m-1 downto (i-1)*m) <= DIN when toleft = '0' and rotate = '0' else
												dout_intern(m-1 downto 0) when toleft = '0' and rotate = '1' else
												dout_intern((i-1)*m-1 downto (i-2)*m);
		end generate;
		
		-- middle: choose between out(i+1) and out(i-1)
		in_mr: if i > 1 and i < n generate
			din_intern(i*m-1 downto (i-1)*m) <= dout_intern((i+1)*m-1 downto i*m) when toleft = '0' else
												dout_intern((i-1)*m-1 downto (i-2)*m);
		end generate;
		
		-- right most: choose between DIN, out(n), out(i+1)
		in_rmr: if i = 1 generate
			din_intern(i*m-1 downto (i-1)*m) <= DIN when toleft = '1' and rotate = '0' else
												dout_intern(n*m-1 downto (n-1)*m) when toleft = '1' and rotate = '1' else
												dout_intern((i+1)*m-1 downto i*m);
		end generate;
		
	end generate;
	
	-- map internal signals to output signals
	DOUT_F <= dout_intern;
	DOUT <= dout_intern(m-1 downto 0) when toleft = '0' else
			dout_intern(n*m-1 downto (n-1)*m);

end Behavioral;

