----------------------------------------------------------------------------------
--
-- Copyright 2019 Ruhr University Bochum, Horst Görtz Institute for IT Security
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy of this software
-- and associated documentation files (the "Software"), to deal in the Software without restriction, 
-- including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, 
-- and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do 
-- so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in all copies or substantial 
-- portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT 
-- LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
-- IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
-- WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
-- SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;


entity pc_uart_receive is
    generic ( fpga_frequency : integer := 100000000;
	   bit_rate : integer := 128000);
    Port ( 	clk : in  STD_LOGIC;
				reset: in STD_LOGIC;
				rs232_bit : in STD_LOGIC;
				data_out : out std_logic_vector(7 downto 0);
				data_valid : out STD_LOGIC
			);
end pc_uart_receive;

architecture Behavioral of pc_uart_receive is

	--  n x m bit shift register with rotation support
	component nxmBitShiftReg is
		generic (
			N		: integer range 2 to 256 := 4;							-- number of registers, IMPORTANT: N = 1 is not allowed (multiplexer config)
			M 		: integer range 1 to 256 := 2							-- bits per register
		);
		port (
			CLK		: in std_logic;											-- input clock
			SR		: in std_logic;											-- set/reset signal
			SRINIT	: in std_logic_vector(n*m-1 downto 0);					-- init value for reset
			CE		: in std_logic;											-- enable signal
			OPMODE	: in std_logic_vector(1 downto 0);						-- operation mode (see list of modes)
			DIN		: in std_logic_vector(m-1 downto 0);					-- input value when shifting
			DOUT	: out std_logic_vector(m-1 downto 0);					-- output: word shifted out resp. last word rotated
			DOUT_F	: out std_logic_vector(n*m-1 downto 0)					-- output: complete register
		);
	end component nxmBitShiftReg;

signal bit_cnt : unsigned(3 downto 0);
signal bit_sampled : std_logic;
signal bit_sampled_v : std_logic_vector(0 downto 0);
signal bit_sampled_enable : std_logic;
signal clk_cnt : integer range 0 to 1048575;
signal data_out_sr : std_logic_vector(8 downto 0);
signal rs232_bit_syn, rs232_bit_last : std_logic;
signal bit_receiving_time : unsigned(19 downto 0);

	-- internal state machine states
	type state_type is (WAIT_DATA, SAMPLING, READY);
	signal state : state_type;

begin

	bit_sampled_v(0) <= bit_sampled;

-- This shiftregister stores the transmitted 9 bits (s12345678)
	Inst_shiftregister: nxmBitShiftReg 
		generic map (
			N => 9,
			M => 1
		)
		port map (
			clk => clk,
			sr => reset,
			srinit => (others => '0'),
			ce => bit_sampled_enable,
			opmode => "00",
			din => bit_sampled_v,
			dout => open,
			dout_f => data_out_sr
		);

-- set duration of 1 bit in fpga-clks
bit_receiving_time <= (to_unsigned(fpga_frequency/bit_rate,bit_receiving_time'length));

-- route the shift register output to top level modul
data_out <=  data_out_sr(8 downto 1);


RECEIVE : process(clk)
	begin
		if rising_edge(clk) then
			if (reset = '1') then
				state <= WAIT_DATA;
				clk_cnt  <= 0;
				bit_cnt <= (others => '0');
				data_valid <= '0';
				bit_sampled <= '0';
				bit_sampled_enable <= '0';
			else
	
				case state is							
					
					when WAIT_DATA =>
											data_valid <= '0';
										-- a rs232_bit falling edge initiates sampling
											if ((rs232_bit_syn = '0') and (rs232_bit_last = '1')) then
												state <= SAMPLING;
											end if;

					when SAMPLING =>
										-- sample value at the middle bit position
											if ((clk_cnt = to_integer(unsigned(bit_receiving_time))/2)) then
												bit_sampled <= rs232_bit;
												bit_sampled_enable <= '1';
											else
												bit_sampled_enable <= '0';
											end if;

										-- Clock counter
											if (clk_cnt < to_integer(bit_receiving_time)) then 
												clk_cnt <= clk_cnt + 1;
											else
												clk_cnt <= 0;
												bit_cnt <= bit_cnt + 1;
											end if;
										
										-- After last bit is sampled, the state is switched to ready
											if (bit_cnt > 8) then
												state <= READY;
											end if;
					
					when READY =>
										-- In Ready-state, the byte is sent
											data_valid <= '1';
											bit_cnt <= (others => '0');
											clk_cnt <= 0;
											state <= WAIT_DATA;
											
				end case;
			end if;
		end if;
	end process;

-- This process buffers signals to realize edge detection
BUFFERING : process(CLK)
	begin
		if rising_edge(CLK) then
			if RESET = '1' then
				rs232_bit_syn <= '0';
				rs232_bit_last <= '0';
			else
				rs232_bit_syn <= rs232_bit;
				rs232_bit_last <= rs232_bit_syn;
			end if;
		end if;
end process;
 
end Behavioral;	