----------------------------------------------------------------------------------
--
-- Copyright 2019 Ruhr University Bochum, Horst Görtz Institute for IT Security
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy of this software
-- and associated documentation files (the "Software"), to deal in the Software without restriction, 
-- including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, 
-- and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do 
-- so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in all copies or substantial 
-- portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT 
-- LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
-- IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
-- WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
-- SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;

entity pc_uart_send is
    generic ( fpga_frequency : integer := 100000000;
	   bit_rate : integer := 128000);
    Port ( clk : in  STD_LOGIC;
			  reset: in STD_LOGIC;
			  data_in : in STD_LOGIC_VECTOR(7 downto 0);
			  rs232_bit : out STD_LOGIC;
			  send_enable: in STD_LOGIC;
			  ready_for_data : out STD_LOGIC);
end pc_uart_send;

architecture Behavioral of pc_uart_send is

	
-- internal state machine states
	type state_type is (READY, SENDING);
	signal state : state_type := READY;

	signal bit_sending_time : unsigned(19 downto 0);
	signal bit_sending_cnt : unsigned(19 downto 0);
	signal bit_cnt : unsigned(3 downto 0);
	signal data_int : std_logic_vector(7 downto 0);
	
begin

-- set duration of 1 bit in fpga-clks
bit_sending_time <= (to_unsigned(fpga_frequency/bit_rate,bit_sending_time'length));

	SEND : process(clk)
	begin
			if rising_edge(clk) then
				if (reset = '1') then
					state <= READY;
					bit_sending_cnt <= (others => '0');
					bit_cnt <= (others => '0');
					rs232_bit <= '1';
					data_int <= data_in;
					ready_for_data <= '1';					
				else
				
				case state is							
					when READY =>
								-- In READY state defaults values are set and the byte is sent when send_enable = 1
									if (send_enable = '1') then
										state <= SENDING;
										ready_for_data <= '0';
									else
										ready_for_data <= '1';
									end if;
									
									bit_sending_cnt <= (others => '0');
									bit_cnt <= (others => '0');
									data_int <= data_in;
									
					when SENDING =>
								--start bit
									if (bit_cnt = 0) then
										rs232_bit <= '0';
								--data bits
									elsif (bit_cnt = 1) then
										rs232_bit <= data_int(0);
									elsif (bit_cnt = 2) then
										rs232_bit <= data_int(1);
									elsif (bit_cnt = 3) then
										rs232_bit <= data_int(2);
									elsif (bit_cnt = 4) then
										rs232_bit <= data_int(3);
									elsif (bit_cnt = 5) then
										rs232_bit <= data_int(4);
									elsif (bit_cnt = 6) then
										rs232_bit <= data_int(5);
									elsif (bit_cnt = 7) then
										rs232_bit <= data_int(6);
									elsif (bit_cnt = 8) then
										rs232_bit <= data_int(7);
									elsif (bit_cnt = 9) then
								-- stop bit
										rs232_bit <= '1';
									else
										state <= READY;
										ready_for_data <= '1';
									end if;
				
								-- UART clock counter and bit selecter
								-- bit_sending_cnt counts from 0 to bit_sending_time,
								-- bit_cnt counts from 0 to 9 (s12345678)
									if (bit_sending_time <= bit_sending_cnt) then
										bit_sending_cnt <= (others => '0');
										bit_cnt <= bit_cnt + 1;
									else
										bit_sending_cnt <= bit_sending_cnt + 1;
										bit_cnt <= bit_cnt + 0;
									end if;
				end case;
			end if;
		end if;
	end process;

end Behavioral;